#!/bin/sh

# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

#
# Run a system test.
#

. ./conf.sh

test $# -gt 0 || {
  echo "usage: $0 test-directory" >&2
  exit 1
}

test=$1
shift

test -d $test || {
  echo "$0: $test: no such test" >&2
  exit 1
}

# Set up any dynamically generated test data
if test -f $test/setup.sh; then
  (cd $test && $SHELL setup.sh "$@")
fi
